@@ -1,3 +1,5 @@
+# vim: ts=8:sts=8:sw=8:noexpandtab
+#
 # This file is part of ReText
 # Copyright: 2012-2015 Dmitry Shachnev
 #
@@ -18,8 +20,8 @@ import markups
 import sys
 from subprocess import Popen
 from ReText import icon_path, app_version, globalSettings, readListFromSettings, \
- writeListToSettings, writeToSettings, datadirs, enchant, enchant_available
-from ReText.tab import ReTextTab, PreviewNormal, PreviewLive
+ writeListToSettings, writeToSettings, datadirs
+from ReText.tab import ReTextTab, ReTextWebPreview, PreviewNormal, PreviewLive
 from ReText.dialogs import HtmlDialog, LocaleDialog
 from ReText.config import ConfigDialog
 from ReText.icontheme import get_icon_theme
@@ -29,13 +31,18 @@ try:
 except ImportError:
 	ReTextFakeVimHandler = None
 
+try:
+	import enchant
+except ImportError:
+	enchant = None
+
 from PyQt5.QtCore import QDir, QFile, QFileInfo, QFileSystemWatcher, \
- QIODevice, QLocale, QRect, QTextCodec, QTextStream, QTimer, QUrl, Qt
+ QIODevice, QLocale, QTextCodec, QTextStream, QTimer, QUrl, Qt
 from PyQt5.QtGui import QColor, QDesktopServices, QIcon, \
- QKeySequence, QPalette, QTextCursor, QTextDocument, QTextDocumentWriter
+ QKeySequence, QPalette, QTextDocument, QTextDocumentWriter
 from PyQt5.QtWidgets import QAction, QActionGroup, QApplication, QCheckBox, \
  QComboBox, QDesktopWidget, QDialog, QFileDialog, QFontDialog, QInputDialog, \
- QLineEdit, QMainWindow, QMenu, QMenuBar, QMessageBox, QTabWidget, QToolBar
+ QLineEdit, QMainWindow, QMenu, QMessageBox, QTabWidget, QToolBar
 from PyQt5.QtPrintSupport import QPrintDialog, QPrintPreviewDialog, QPrinter
 
 class ReTextWindow(QMainWindow):
@@ -49,6 +56,11 @@ class ReTextWindow(QMainWindow):
 			self.move((screenRect.width()-self.width())/2, (screenRect.height()-self.height())/2)
 		if not screenRect.contains(self.geometry()):
 			self.showMaximized()
+		if sys.platform.startswith('darwin'):
+			# https://github.com/retext-project/retext/issues/198
+			searchPaths = QIcon.themeSearchPaths()
+			searchPaths.append('/opt/local/share/icons')
+			QIcon.setThemeSearchPaths(searchPaths)
 		if globalSettings.iconTheme:
 			QIcon.setThemeName(globalSettings.iconTheme)
 		if QIcon.themeName() in ('hicolor', ''):
@@ -125,7 +137,8 @@ class ReTextWindow(QMainWindow):
 		menuPreview = QMenu()
 		menuPreview.addAction(self.actionLivePreview)
 		self.actionPreview.setMenu(menuPreview)
-		self.actionTableMode = self.act(self.tr('Table mode'), shct=Qt.CTRL+Qt.Key_T,
+		self.actionTableMode = self.act(self.tr('Table editing mode'),
+			shct=Qt.CTRL+Qt.Key_T,
 			trigbool=lambda x: self.currentTab.editBox.enableTableMode(x))
 		if ReTextFakeVimHandler:
 			self.actionFakeVimMode = self.act(self.tr('FakeVim mode'),
@@ -163,16 +176,22 @@ class ReTextWindow(QMainWindow):
 		qApp = QApplication.instance()
 		qApp.clipboard().dataChanged.connect(self.clipboardDataChanged)
 		self.clipboardDataChanged()
-		if enchant_available:
+		if enchant is not None:
 			self.actionEnableSC = self.act(self.tr('Enable'), trigbool=self.enableSpellCheck)
 			self.actionSetLocale = self.act(self.tr('Set locale'), trig=self.changeLocale)
 		self.actionWebKit = self.act(self.tr('Use WebKit renderer'), trigbool=self.enableWebKit)
+		if ReTextWebPreview is None:
+			globalSettings.useWebKit = False
+			self.actionWebKit.setEnabled(False)
 		self.actionWebKit.setChecked(globalSettings.useWebKit)
 		self.actionShow = self.act(self.tr('Show directory'), 'system-file-manager', self.showInDir)
 		self.actionFind = self.act(self.tr('Next'), 'go-next', self.find,
 			shct=QKeySequence.FindNext)
 		self.actionFindPrev = self.act(self.tr('Previous'), 'go-previous',
 			lambda: self.find(back=True), shct=QKeySequence.FindPrevious)
+		self.actionCloseSearch = self.act(self.tr('Close'), 'window-close',
+			lambda: self.searchBar.setVisible(False))
+		self.actionCloseSearch.setPriority(QAction.LowPriority)
 		self.actionHelp = self.act(self.tr('Get help online'), 'help-contents', self.openHelp)
 		self.aboutWindowTitle = self.tr('About ReText')
 		self.actionAbout = self.act(self.aboutWindowTitle, 'help-about', self.aboutDialog)
@@ -183,11 +202,12 @@ class ReTextWindow(QMainWindow):
 		availableMarkups = markups.get_available_markups()
 		if not availableMarkups:
 			print('Warning: no markups are available!')
-		self.defaultMarkup = availableMarkups[0] if availableMarkups else None
+		defaultMarkup = availableMarkups[0] if availableMarkups else None
 		if globalSettings.defaultMarkup:
 			mc = markups.find_markup_class_by_name(globalSettings.defaultMarkup)
 			if mc and mc.available():
-				self.defaultMarkup = mc
+				defaultMarkup = mc
+		self.setDefaultMarkup(defaultMarkup)
 		if len(availableMarkups) > 1:
 			self.chooseGroup = QActionGroup(self)
 			markupActions = []
@@ -198,28 +218,26 @@ class ReTextWindow(QMainWindow):
 				self.chooseGroup.addAction(markupAction)
 				markupActions.append(markupAction)
 		self.actionBold = self.act(self.tr('Bold'), shct=QKeySequence.Bold,
-			trig=lambda: self.insertChars('**'))
+			trig=lambda: self.insertFormatting('bold'))
 		self.actionItalic = self.act(self.tr('Italic'), shct=QKeySequence.Italic,
-			trig=lambda: self.insertChars('*'))
+			trig=lambda: self.insertFormatting('italic'))
 		self.actionUnderline = self.act(self.tr('Underline'), shct=QKeySequence.Underline,
-			trig=lambda: self.insertTag('u'))
-		self.usefulTags = ('a', 'big', 'center', 'img', 's', 'small', 'span',
-			'table', 'td', 'tr', 'u')
+			trig=lambda: self.insertFormatting('underline'))
+		self.usefulTags = ('header', 'italic', 'bold', 'underline', 'numbering',
+			'bullets', 'image', 'link', 'inline code', 'code block', 'blockquote')
 		self.usefulChars = ('deg', 'divide', 'dollar', 'hellip', 'laquo', 'larr',
 			'lsquo', 'mdash', 'middot', 'minus', 'nbsp', 'ndash', 'raquo',
 			'rarr', 'rsquo', 'times')
-		self.tagsBox = QComboBox(self.editBar)
-		self.tagsBox.addItem(self.tr('Tags'))
-		self.tagsBox.addItems(self.usefulTags)
-		self.tagsBox.activated.connect(self.insertTag)
+		self.formattingBox = QComboBox(self.editBar)
+		self.formattingBox.addItem(self.tr('Formatting'))
+		self.formattingBox.addItems(self.usefulTags)
+		self.formattingBox.activated[str].connect(self.insertFormatting)
 		self.symbolBox = QComboBox(self.editBar)
 		self.symbolBox.addItem(self.tr('Symbols'))
 		self.symbolBox.addItems(self.usefulChars)
 		self.symbolBox.activated.connect(self.insertSymbol)
 		self.updateStyleSheet()
-		menubar = QMenuBar(self)
-		menubar.setGeometry(QRect(0, 0, 800, 25))
-		self.setMenuBar(menubar)
+		menubar = self.menuBar()
 		menuFile = menubar.addMenu(self.tr('File'))
 		menuEdit = menubar.addMenu(self.tr('Edit'))
 		menuHelp = menubar.addMenu(self.tr('Help'))
@@ -227,7 +245,6 @@ class ReTextWindow(QMainWindow):
 		menuFile.addAction(self.actionOpen)
 		self.menuRecentFiles = menuFile.addMenu(self.tr('Open recent'))
 		self.menuRecentFiles.aboutToShow.connect(self.updateRecentFiles)
-		menuFile.addMenu(self.menuRecentFiles)
 		menuFile.addAction(self.actionShow)
 		menuFile.addAction(self.actionSetEncoding)
 		menuFile.addAction(self.actionReload)
@@ -258,7 +275,7 @@ class ReTextWindow(QMainWindow):
 		menuEdit.addAction(self.actionCopy)
 		menuEdit.addAction(self.actionPaste)
 		menuEdit.addSeparator()
-		if enchant_available:
+		if enchant is not None:
 			menuSC = menuEdit.addMenu(self.tr('Spell check'))
 			menuSC.addAction(self.actionEnableSC)
 			menuSC.addAction(self.actionSetLocale)
@@ -288,9 +305,6 @@ class ReTextWindow(QMainWindow):
 		menuHelp.addSeparator()
 		menuHelp.addAction(self.actionAbout)
 		menuHelp.addAction(self.actionAboutQt)
-		menubar.addMenu(menuFile)
-		menubar.addMenu(menuEdit)
-		menubar.addMenu(menuHelp)
 		toolBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
 		toolBar.addAction(self.actionNew)
 		toolBar.addSeparator()
@@ -307,7 +321,7 @@ class ReTextWindow(QMainWindow):
 		self.editBar.addAction(self.actionCopy)
 		self.editBar.addAction(self.actionPaste)
 		self.editBar.addSeparator()
-		self.editBar.addWidget(self.tagsBox)
+		self.editBar.addWidget(self.formattingBox)
 		self.editBar.addWidget(self.symbolBox)
 		self.searchEdit = QLineEdit(self.searchBar)
 		self.searchEdit.setPlaceholderText(self.tr('Search'))
@@ -318,6 +332,7 @@ class ReTextWindow(QMainWindow):
 		self.searchBar.addWidget(self.csBox)
 		self.searchBar.addAction(self.actionFindPrev)
 		self.searchBar.addAction(self.actionFind)
+		self.searchBar.addAction(self.actionCloseSearch)
 		self.searchBar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
 		self.searchBar.setVisible(False)
 		self.autoSaveEnabled = globalSettings.autoSave
@@ -326,23 +341,21 @@ class ReTextWindow(QMainWindow):
 			timer.start(60000)
 			timer.timeout.connect(self.saveAll)
 		self.ind = None
-		if enchant_available:
+		if enchant is not None:
 			self.sl = globalSettings.spellCheckLocale
-			if self.sl:
-				try:
-					enchant.Dict(self.sl)
-				except Exception as e:
-					print(e, file=sys.stderr)
-					self.sl = None
+			try:
+				enchant.Dict(self.sl or None)
+			except enchant.errors.Error as e:
+				print(e, file=sys.stderr)
+				globalSettings.spellCheck = False
 			if globalSettings.spellCheck:
 				self.actionEnableSC.setChecked(True)
-				self.enableSpellCheck(True)
 		self.fileSystemWatcher = QFileSystemWatcher()
 		self.fileSystemWatcher.fileChanged.connect(self.fileChanged)
 
 	def iterateTabs(self):
 		for i in range(self.tabWidget.count()):
-			yield self.tabWidget.widget(i).tab
+			yield self.tabWidget.widget(i)
 
 	def updateStyleSheet(self):
 		if globalSettings.styleSheet:
@@ -351,7 +364,10 @@ class ReTextWindow(QMainWindow):
 			self.ss = QTextStream(sheetfile).readAll()
 			sheetfile.close()
 		else:
-			self.ss = ''
+			palette = QApplication.palette()
+			self.ss = 'html { color: %s; }\n' % palette.color(QPalette.WindowText).name()
+			self.ss += 'td, th { border: 1px solid #c3c3c3; padding: 0 3px 0 3px; }\n'
+			self.ss += 'table { border-collapse: collapse; }\n'
 
 	def initTabWidget(self):
 		def dragEnterEvent(e):
@@ -388,41 +404,89 @@ class ReTextWindow(QMainWindow):
 
 	def printError(self):
 		import traceback
-		print('Exception occured while parsing document:', file=sys.stderr)
+		print('Exception occurred while parsing document:', file=sys.stderr)
 		traceback.print_exc()
 
+
+	def tabFileNameChanged(self, tab):
+		'''
+		Perform all UI state changes that need to be done when the
+		filename of the current tab has changed.
+		'''
+		if tab == self.currentTab:
+			if tab.fileName:
+				self.setWindowTitle("")
+				self.setWindowFilePath(tab.fileName)
+				self.tabWidget.setTabText(self.ind, tab.getBaseName())
+				self.tabWidget.setTabToolTip(self.ind, tab.fileName)
+				QDir.setCurrent(QFileInfo(tab.fileName).dir().path())
+			else:
+				self.setWindowTitle(self.tr('New document') + '[*]')
+
+			canReload = bool(tab.fileName) and not self.autoSaveActive(tab)
+			self.actionSetEncoding.setEnabled(canReload)
+			self.actionReload.setEnabled(canReload)
+
+	def tabActiveMarkupChanged(self, tab):
+		'''
+		Perform all UI state changes that need to be done when the
+		active markup class of the current tab has changed.
+		'''
+		if tab == self.currentTab:
+			markupClass = tab.getActiveMarkupClass()
+			dtMarkdown = (markupClass == markups.MarkdownMarkup)
+			dtMkdOrReST = dtMarkdown or (markupClass == markups.ReStructuredTextMarkup)
+			self.formattingBox.setEnabled(dtMarkdown)
+			self.symbolBox.setEnabled(dtMarkdown)
+			self.actionUnderline.setEnabled(dtMarkdown)
+			self.actionBold.setEnabled(dtMkdOrReST)
+			self.actionItalic.setEnabled(dtMkdOrReST)
+
+	def tabModificationStateChanged(self, tab):
+		'''
+		Perform all UI state changes that need to be done when the
+		modification state of the current tab has changed.
+		'''
+		if tab == self.currentTab:
+			changed = tab.editBox.document().isModified()
+			if self.autoSaveActive(tab):
+				changed = False
+			self.actionSave.setEnabled(changed)
+			self.setWindowModified(changed)
+
 	def createTab(self, fileName):
-		self.currentTab = ReTextTab(self, fileName)
-		self.tabWidget.addTab(self.currentTab.getSplitter(), self.tr("New document"))
+		self.currentTab = ReTextTab(self, fileName, self.defaultMarkup,
+			previewState=int(globalSettings.livePreviewByDefault))
+		self.currentTab.fileNameChanged.connect(lambda: self.tabFileNameChanged(self.currentTab))
+		self.currentTab.modificationStateChanged.connect(lambda: self.tabModificationStateChanged(self.currentTab))
+		self.currentTab.activeMarkupChanged.connect(lambda: self.tabActiveMarkupChanged(self.currentTab))
+		self.tabWidget.addTab(self.currentTab, self.tr("New document"))
+		self.currentTab.updateBoxesVisibility()
 
 	def closeTab(self, ind):
 		if self.maybeSave(ind):
 			if self.tabWidget.count() == 1:
 				self.createTab("")
-			currentWidget = self.tabWidget.widget(ind)
-			if currentWidget.tab.fileName:
-				self.fileSystemWatcher.removePath(currentWidget.tab.fileName)
-			del currentWidget.tab
+			closedTab = self.tabWidget.widget(ind)
+			if closedTab.fileName:
+				self.fileSystemWatcher.removePath(closedTab.fileName)
 			self.tabWidget.removeTab(ind)
-
-	def docTypeChanged(self):
-		markupClass = self.currentTab.getMarkupClass()
-		if type(self.currentTab.markup) != markupClass:
-			self.currentTab.setMarkupClass(markupClass)
-			self.currentTab.updatePreviewBox()
-		dtMarkdown = (markupClass == markups.MarkdownMarkup)
-		dtMkdOrReST = dtMarkdown or (markupClass == markups.ReStructuredTextMarkup)
-		self.tagsBox.setEnabled(dtMarkdown)
-		self.symbolBox.setEnabled(dtMarkdown)
-		self.actionUnderline.setEnabled(dtMarkdown)
-		self.actionBold.setEnabled(dtMkdOrReST)
-		self.actionItalic.setEnabled(dtMkdOrReST)
-		canReload = bool(self.currentTab.fileName) and not self.autoSaveActive()
-		self.actionSetEncoding.setEnabled(canReload)
-		self.actionReload.setEnabled(canReload)
+			closedTab.deleteLater()
 
 	def changeIndex(self, ind):
-		self.currentTab = self.tabWidget.currentWidget().tab
+		'''
+		This function is called when a different tab is selected.
+		It changes the state of the window to mirror the current state
+		of the newly selected tab. Future changes to this state will be
+		done in response to signals emitted by the tab, to which the
+		window was subscribed when the tab was created. The window is
+		subscribed to all tabs like this, but only the active tab will
+		logically generate these signals.
+		Aside from the above this function also calls the handlers for
+		the other changes that are implied by a tab switch: filename
+		change, modification state change and active markup change.
+		'''
+		self.currentTab = self.tabWidget.currentWidget()
 		editBox = self.currentTab.editBox
 		previewState = self.currentTab.previewState
 		self.actionUndo.setEnabled(editBox.document().isUndoAvailable())
@@ -434,16 +498,12 @@ class ReTextWindow(QMainWindow):
 		self.actionTableMode.setChecked(editBox.tableModeEnabled)
 		self.editBar.setEnabled(previewState < PreviewNormal)
 		self.ind = ind
-		if self.currentTab.fileName:
-			self.setCurrentFile()
-		else:
-			self.setWindowTitle(self.tr('New document') + '[*]')
-			self.docTypeChanged()
-		self.modificationChanged(editBox.document().isModified())
-		if globalSettings.restorePreviewState:
-			globalSettings.previewState = (previewState >= PreviewLive)
 		editBox.setFocus(Qt.OtherFocusReason)
 
+		self.tabFileNameChanged(self.currentTab)
+		self.tabModificationStateChanged(self.currentTab)
+		self.tabActiveMarkupChanged(self.currentTab)
+
 	def changeEditorFont(self):
 		font, ok = QFontDialog.getFont(globalSettings.editorFont, self)
 		if ok:
@@ -456,38 +516,31 @@ class ReTextWindow(QMainWindow):
 		if ok:
 			globalSettings.font = font
 			for tab in self.iterateTabs():
-				tab.updatePreviewBox()
+				tab.triggerPreviewUpdate()
 
 	def preview(self, viewmode):
 		self.currentTab.previewState = viewmode * 2
 		self.actionLivePreview.setChecked(False)
 		self.editBar.setDisabled(viewmode)
 		self.currentTab.updateBoxesVisibility()
-		if viewmode:
-			self.currentTab.updatePreviewBox()
 
 	def enableLivePreview(self, livemode):
-		if globalSettings.restorePreviewState:
-			globalSettings.previewState = livemode
 		self.currentTab.previewState = int(livemode)
 		self.actionPreview.setChecked(livemode)
 		self.editBar.setEnabled(True)
 		self.currentTab.updateBoxesVisibility()
-		if livemode:
-			self.currentTab.updatePreviewBox()
 
 	def enableWebKit(self, enable):
 		globalSettings.useWebKit = enable
-		for i in range(self.tabWidget.count()):
-			splitter = self.tabWidget.widget(i)
-			tab = splitter.tab
+		for tab in self.iterateTabs():
+			tab.previewBox.disconnectExternalSignals()
 			tab.previewBox.setParent(None)
 			tab.previewBox.deleteLater()
-			tab.previewBox = tab.createPreviewBox()
+			tab.previewBox = tab.createPreviewBox(tab.editBox)
 			tab.previewBox.setMinimumWidth(125)
-			splitter.addWidget(tab.previewBox)
-			splitter.setSizes((50, 50))
-			tab.updatePreviewBox()
+			tab.addWidget(tab.previewBox)
+			tab.setSizes((50, 50))
+			tab.triggerPreviewUpdate()
 			tab.updateBoxesVisibility()
 
 	def enableCopy(self, copymode):
@@ -515,13 +568,13 @@ class ReTextWindow(QMainWindow):
 			FakeVimMode.exit(self)
 
 	def enableSpellCheck(self, yes):
-		if yes:
-			if self.sl:
-				self.setAllDictionaries(enchant.Dict(self.sl))
-			else:
-				self.setAllDictionaries(enchant.Dict())
-		else:
-			self.setAllDictionaries(None)
+		try:
+			dict = enchant.Dict(self.sl or None)
+		except enchant.errors.Error as e:
+			QMessageBox.warning(self, '', str(e))
+			self.actionEnableSC.setChecked(False)
+			yes = False
+		self.setAllDictionaries(dict if yes else None)
 		globalSettings.spellCheck = yes
 
 	def setAllDictionaries(self, dictionary):
@@ -531,28 +584,19 @@ class ReTextWindow(QMainWindow):
 			hl.rehighlight()
 
 	def changeLocale(self):
-		if self.sl:
-			localedlg = LocaleDialog(self, defaultText=self.sl)
-		else:
-			localedlg = LocaleDialog(self)
+		localedlg = LocaleDialog(self, defaultText=self.sl)
 		if localedlg.exec() != QDialog.Accepted:
 			return
 		sl = localedlg.localeEdit.text()
-		setdefault = localedlg.checkBox.isChecked()
-		if sl:
-			try:
-				sl = str(sl)
-				enchant.Dict(sl)
-			except Exception as e:
-				QMessageBox.warning(self, '', str(e))
-			else:
-				self.sl = sl
-				self.enableSpellCheck(self.actionEnableSC.isChecked())
+		try:
+			enchant.Dict(sl or None)
+		except enchant.errors.Error as e:
+			QMessageBox.warning(self, '', str(e))
 		else:
-			self.sl = None
+			self.sl = sl or None
 			self.enableSpellCheck(self.actionEnableSC.isChecked())
-		if setdefault:
-			globalSettings.spellCheckLocale = sl
+			if localedlg.checkBox.isChecked():
+				globalSettings.spellCheckLocale = sl
 
 	def searchBarVisibilityChanged(self, visible):
 		self.actionSearch.setChecked(visible)
@@ -566,18 +610,8 @@ class ReTextWindow(QMainWindow):
 		if self.csBox.isChecked():
 			flags |= QTextDocument.FindCaseSensitively
 		text = self.searchEdit.text()
-		editBox = self.currentTab.editBox
-		cursor = editBox.textCursor()
-		newCursor = editBox.document().find(text, cursor, flags)
-		if not newCursor.isNull():
-			editBox.setTextCursor(newCursor)
-			return self.setSearchEditColor(True)
-		cursor.movePosition(QTextCursor.End if back else QTextCursor.Start)
-		newCursor = editBox.document().find(text, cursor, flags)
-		if not newCursor.isNull():
-			editBox.setTextCursor(newCursor)
-			return self.setSearchEditColor(True)
-		self.setSearchEditColor(False)
+		found = self.currentTab.find(text, flags)
+		self.setSearchEditColor(found)
 
 	def setSearchEditColor(self, found):
 		palette = self.searchEdit.palette()
@@ -592,19 +626,15 @@ class ReTextWindow(QMainWindow):
 		else:
 			QMessageBox.warning(self, '', self.tr("Please, save the file somewhere."))
 
-	def setCurrentFile(self):
-		self.setWindowTitle("")
-		self.tabWidget.setTabText(self.ind, self.currentTab.getDocumentTitle(baseName=True))
-		self.setWindowFilePath(self.currentTab.fileName)
-		files = readListFromSettings("recentFileList")
-		while self.currentTab.fileName in files:
-			files.remove(self.currentTab.fileName)
-		files.insert(0, self.currentTab.fileName)
-		if len(files) > 10:
-			del files[10:]
-		writeListToSettings("recentFileList", files)
-		QDir.setCurrent(QFileInfo(self.currentTab.fileName).dir().path())
-		self.docTypeChanged()
+	def moveToTopOfRecentFileList(self, fileName):
+		if fileName:
+			files = readListFromSettings("recentFileList")
+			if fileName in files:
+				files.remove(fileName)
+			files.insert(0, fileName)
+			if len(files) > 10:
+				del files[10:]
+			writeListToSettings("recentFileList", files)
 
 	def createNew(self, text=None):
 		self.createTab("")
@@ -635,7 +665,7 @@ class ReTextWindow(QMainWindow):
 	def openFunction(self, fileName):
 		return lambda: self.openFileWrapper(fileName)
 
-	def extensionFuntion(self, data):
+	def extensionFunction(self, data):
 		return lambda: \
 		self.runExtensionCommand(data['Exec'], data['FileFilter'], data['DefaultExtension'])
 
@@ -665,7 +695,7 @@ class ReTextWindow(QMainWindow):
 						data[prop] = extension[prop]
 					else:
 						data[prop] = ''
-				action = self.act(name, trig=self.extensionFuntion(data))
+				action = self.act(name, trig=self.extensionFunction(data))
 				if 'Icon' in extension:
 					action.setIcon(self.actIcon(extension['Icon']))
 				mimetype = extension['MimeType'] if 'MimeType' in extension else None
@@ -675,13 +705,13 @@ class ReTextWindow(QMainWindow):
 				self.extensionActions.append((action, mimetype))
 
 	def updateExtensionsVisibility(self):
-		markupClass = self.currentTab.getMarkupClass()
+		markupClass = self.currentTab.getActiveMarkupClass()
 		for action in self.extensionActions:
 			if markupClass is None:
 				action[0].setEnabled(False)
 				continue
 			mimetype = action[1]
-			if mimetype == None:
+			if mimetype is None:
 				enabled = True
 			elif markupClass == markups.MarkdownMarkup:
 				enabled = (mimetype in ("text/x-retext-markdown", "text/x-markdown"))
@@ -739,11 +769,8 @@ class ReTextWindow(QMainWindow):
 				self.tabWidget.setCurrentIndex(self.ind)
 			if fileName:
 				self.fileSystemWatcher.addPath(fileName)
-			self.currentTab.fileName = fileName
-			self.currentTab.readTextFromFile()
-			editBox = self.currentTab.editBox
-			self.setCurrentFile()
-			self.setWindowModified(editBox.document().isModified())
+			self.currentTab.readTextFromFile(fileName)
+			self.moveToTopOfRecentFileList(self.currentTab.fileName)
 
 	def showEncodingDialog(self):
 		if not self.maybeSave(self.ind):
@@ -753,7 +780,7 @@ class ReTextWindow(QMainWindow):
 			[bytes(b).decode() for b in QTextCodec.availableCodecs()],
 			0, False)
 		if ok:
-			self.currentTab.readTextFromFile(encoding)
+			self.currentTab.readTextFromFile(None, encoding)
 
 	def saveFileAs(self):
 		self.saveFile(dlg=True)
@@ -762,11 +789,12 @@ class ReTextWindow(QMainWindow):
 		for tab in self.iterateTabs():
 			if tab.fileName and QFileInfo(tab.fileName).isWritable():
 				tab.saveTextToFile()
-				tab.editBox.document().setModified(False)
 
 	def saveFile(self, dlg=False):
-		if (not self.currentTab.fileName) or dlg:
-			markupClass = self.currentTab.getMarkupClass()
+		fileNameToSave = self.currentTab.fileName
+
+		if (not fileNameToSave) or dlg:
+			markupClass = self.currentTab.getActiveMarkupClass()
 			if (markupClass is None) or not hasattr(markupClass, 'default_extension'):
 				defaultExt = self.tr("Plain text (*.txt)")
 				ext = ".txt"
@@ -781,20 +809,15 @@ class ReTextWindow(QMainWindow):
 					ext = globalSettings.restDefaultFileExtension
 				else:
 					ext = markupClass.default_extension
-			newFileName = QFileDialog.getSaveFileName(self,
+			fileNameToSave = QFileDialog.getSaveFileName(self,
 				self.tr("Save file"), "", defaultExt)[0]
-			if newFileName:
-				if not QFileInfo(newFileName).suffix():
-					newFileName += ext
-				if self.currentTab.fileName:
-					self.fileSystemWatcher.removePath(self.currentTab.fileName)
-				self.currentTab.fileName = newFileName
+			if fileNameToSave:
+				if not QFileInfo(fileNameToSave).suffix():
+					fileNameToSave += ext
 				self.actionSetEncoding.setDisabled(self.autoSaveActive())
-		if self.currentTab.fileName:
-			if self.currentTab.saveTextToFile():
-				self.setCurrentFile()
-				self.currentTab.editBox.document().setModified(False)
-				self.setWindowModified(False)
+		if fileNameToSave:
+			if self.currentTab.saveTextToFile(fileNameToSave):
+				self.moveToTopOfRecentFileList(self.currentTab.fileName)
 				return True
 			else:
 				QMessageBox.warning(self, '',
@@ -805,8 +828,8 @@ class ReTextWindow(QMainWindow):
 		if not QFileInfo(fileName).suffix():
 			fileName += ".html"
 		try:
-			htmltext = self.currentTab.getHtml(includeStyleSheet=False,
-				includeMeta=True, webenv=True)
+			_, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=False,
+				                                              webenv=True)
 		except Exception:
 			return self.printError()
 		htmlFile = QFile(fileName)
@@ -817,19 +840,20 @@ class ReTextWindow(QMainWindow):
 		html << htmltext
 		htmlFile.close()
 
-	def textDocument(self):
+	def textDocument(self, title, htmltext):
 		td = QTextDocument()
-		td.setMetaInformation(QTextDocument.DocumentTitle,
-		                      self.currentTab.getDocumentTitle())
+		td.setMetaInformation(QTextDocument.DocumentTitle, title)
 		if self.ss:
 			td.setDefaultStyleSheet(self.ss)
-		td.setHtml(self.currentTab.getHtml())
+		td.setHtml(htmltext)
 		td.setDefaultFont(globalSettings.font)
 		return td
 
 	def saveOdf(self):
+		title, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=True,
+		                                                          webenv=False)
 		try:
-			document = self.textDocument()
+			document = self.textDocument(title, htmltext)
 		except Exception:
 			return self.printError()
 		fileName = QFileDialog.getSaveFileName(self,
@@ -848,55 +872,60 @@ class ReTextWindow(QMainWindow):
 		if fileName:
 			self.saveHtml(fileName)
 
-	def getDocumentForPrint(self):
+	def getDocumentForPrint(self, title, htmltext, preview):
 		if globalSettings.useWebKit:
-			return self.currentTab.previewBox
+			return preview
 		try:
-			return self.textDocument()
+			return self.textDocument(title, htmltext)
 		except Exception:
 			self.printError()
 
-	def standardPrinter(self):
+	def standardPrinter(self, title):
 		printer = QPrinter(QPrinter.HighResolution)
-		printer.setDocName(self.currentTab.getDocumentTitle())
+		printer.setDocName(title)
 		printer.setCreator('ReText %s' % app_version)
 		return printer
 
 	def savePdf(self):
-		self.currentTab.updatePreviewBox()
 		fileName = QFileDialog.getSaveFileName(self,
 			self.tr("Export document to PDF"),
 			"", self.tr("PDF files (*.pdf)"))[0]
 		if fileName:
 			if not QFileInfo(fileName).suffix():
 				fileName += ".pdf"
-			printer = self.standardPrinter()
+			title, htmltext, preview = self.currentTab.getDocumentForExport(includeStyleSheet=True,
+										        webenv=False)
+			printer = self.standardPrinter(title)
 			printer.setOutputFormat(QPrinter.PdfFormat)
 			printer.setOutputFileName(fileName)
-			document = self.getDocumentForPrint()
+			document = self.getDocumentForPrint(title, htmltext, preview)
 			if document != None:
 				document.print(printer)
 
 	def printFile(self):
-		self.currentTab.updatePreviewBox()
-		printer = self.standardPrinter()
+		title, htmltext, preview = self.currentTab.getDocumentForExport(includeStyleSheet=True,
+										webenv=False)
+		printer = self.standardPrinter(title)
 		dlg = QPrintDialog(printer, self)
 		dlg.setWindowTitle(self.tr("Print document"))
 		if (dlg.exec() == QDialog.Accepted):
-			document = self.getDocumentForPrint()
+			document = self.getDocumentForPrint(title, htmltext, preview)
 			if document != None:
 				document.print(printer)
 
 	def printPreview(self):
-		document = self.getDocumentForPrint()
-		if document == None:
+		title, htmltext, preview = self.currentTab.getDocumentForExport(includeStyleSheet=True,
+										webenv=False)
+		document = self.getDocumentForPrint(title, htmltext, preview)
+		if document is None:
 			return
-		printer = self.standardPrinter()
+		printer = self.standardPrinter(title)
 		preview = QPrintPreviewDialog(printer, self)
 		preview.paintRequested.connect(document.print)
 		preview.exec()
 
 	def runExtensionCommand(self, command, filefilter, defaultext):
+		import shlex
 		of = ('%of' in command)
 		html = ('%html' in command)
 		if of:
@@ -908,15 +937,17 @@ class ReTextWindow(QMainWindow):
 				return
 			if defaultext and not QFileInfo(fileName).suffix():
 				fileName += defaultext
-		basename = '.%s.retext-temp' % self.currentTab.getDocumentTitle(baseName=True)
+		else:
+			fileName = 'out' + defaultext
+		basename = '.%s.retext-temp' % self.currentTab.getBaseName()
 		if html:
 			tmpname = basename+'.html'
 			self.saveHtml(tmpname)
 		else:
-			tmpname = basename + self.currentTab.getMarkupClass().default_extension
-			self.currentTab.saveTextToFile(fileName=tmpname, addToWatcher=False)
-		command = command.replace('%of', '"out'+defaultext+'"')
-		command = command.replace('%html' if html else '%if', '"'+tmpname+'"')
+			tmpname = basename + self.currentTab.getActiveMarkupClass().default_extension
+			self.currentTab.writeTextToFile(tmpname)
+		command = command.replace('%of', shlex.quote(fileName))
+		command = command.replace('%html' if html else '%if', shlex.quote(tmpname))
 		try:
 			Popen(str(command), shell=True).wait()
 		except Exception as error:
@@ -924,55 +955,62 @@ class ReTextWindow(QMainWindow):
 			QMessageBox.warning(self, '', self.tr('Failed to execute the command:')
 			+ '\n' + errorstr)
 		QFile(tmpname).remove()
-		if of:
-			QFile('out'+defaultext).rename(fileName)
 
-	def autoSaveActive(self, ind=None):
-		tab = self.currentTab if ind is None else self.tabWidget.widget(ind).tab
-		return (self.autoSaveEnabled and tab.fileName and
-			QFileInfo(tab.fileName).isWritable())
-
-	def modificationChanged(self, changed):
-		if self.autoSaveActive():
-			changed = False
-		self.actionSave.setEnabled(changed)
-		self.setWindowModified(changed)
+	def autoSaveActive(self, tab=None):
+		tab = tab if tab else self.currentTab
+		return bool(self.autoSaveEnabled and tab.fileName and
+			    QFileInfo(tab.fileName).isWritable())
 
 	def clipboardDataChanged(self):
 		mimeData = QApplication.instance().clipboard().mimeData()
 		if mimeData is not None:
-			self.actionPaste.setEnabled(mimeData.hasText())
+			self.actionPaste.setEnabled(mimeData.hasText() or mimeData.hasImage())
 
-	def insertChars(self, chars):
-		tc = self.currentTab.editBox.textCursor()
-		if tc.hasSelection():
-			selection = tc.selectedText()
-			if selection.startswith(chars) and selection.endswith(chars):
-				if len(selection) > 2*len(chars):
-					selection = selection[len(chars):-len(chars)]
-					tc.insertText(selection)
-			else:
-				tc.insertText(chars+tc.selectedText()+chars)
-		else:
-			tc.insertText(chars)
+	def insertFormatting(self, formatting):
+		cursor = self.currentTab.editBox.textCursor()
+		text = cursor.selectedText()
+		moveCursorTo = None
 
-	def insertTag(self, ut):
-		if not ut:
+		def c(cursor):
+			nonlocal moveCursorTo
+			moveCursorTo = cursor.position()
+
+		def ensurenl(cursor):
+			if not cursor.atBlockStart():
+				cursor.insertText('\n\n')
+
+		toinsert = {
+			'header': (ensurenl, '# ', text),
+			'italic': ('*', text, c, '*'),
+			'bold': ('**', text, c, '**'),
+			'underline': ('<u>', text, c, '</u>'),
+			'numbering': (ensurenl, ' 1. ', text),
+			'bullets': (ensurenl, '  * ', text),
+			'image': ('![', text or self.tr('Alt text'), c, '](', self.tr('URL'), ')'),
+			'link': ('[', text or self.tr('Link text'), c, '](', self.tr('URL'), ')'),
+			'inline code': ('`', text, c, '`'),
+			'code block': (ensurenl, '    ', text),
+			'blockquote': (ensurenl, '> ', text),
+		}
+
+		if formatting not in toinsert:
 			return
-		if isinstance(ut, int):
-			ut = self.usefulTags[ut - 1]
-		arg = ' style=""' if ut == 'span' else ''
-		tc = self.currentTab.editBox.textCursor()
-		if ut == 'img':
-			toinsert = ('<a href="' + tc.selectedText() +
-			'" target="_blank"><img src="' + tc.selectedText() + '"/></a>')
-		elif ut == 'a':
-			toinsert = ('<a href="' + tc.selectedText() +
-			'" target="_blank">' + tc.selectedText() + '</a>')
-		else:
-			toinsert = '<'+ut+arg+'>'+tc.selectedText()+'</'+ut+'>'
-		tc.insertText(toinsert)
-		self.tagsBox.setCurrentIndex(0)
+
+		cursor.beginEditBlock()
+		for token in toinsert[formatting]:
+			if callable(token):
+				token(cursor)
+			else:
+				cursor.insertText(token)
+		cursor.endEditBlock()
+
+		self.formattingBox.setCurrentIndex(0)
+		# Bring back the focus on the editor
+		self.currentTab.editBox.setFocus(Qt.OtherFocusReason)
+
+		if moveCursorTo:
+			cursor.setPosition(moveCursorTo)
+			self.currentTab.editBox.setTextCursor(cursor)
 
 	def insertSymbol(self, num):
 		if num:
@@ -995,7 +1033,6 @@ class ReTextWindow(QMainWindow):
 		elif not self.currentTab.editBox.document().isModified():
 			# File was not modified in ReText, reload silently
 			self.currentTab.readTextFromFile()
-			self.currentTab.updatePreviewBox()
 		else:
 			text = self.tr(
 				'This document has been modified by other application.\n'
@@ -1011,7 +1048,6 @@ class ReTextWindow(QMainWindow):
 			messageBox.exec()
 			if messageBox.clickedButton() is reloadButton:
 				self.currentTab.readTextFromFile()
-				self.currentTab.updatePreviewBox()
 			else:
 				self.autoSaveEnabled = False
 				self.currentTab.editBox.document().setModified(True)
@@ -1020,8 +1056,8 @@ class ReTextWindow(QMainWindow):
 			self.fileSystemWatcher.addPath(fileName)
 
 	def maybeSave(self, ind):
-		tab = self.tabWidget.widget(ind).tab
-		if self.autoSaveActive(ind):
+		tab = self.tabWidget.widget(ind)
+		if self.autoSaveActive(tab):
 			tab.saveTextToFile()
 			return True
 		if not tab.editBox.document().isModified():
@@ -1047,11 +1083,11 @@ class ReTextWindow(QMainWindow):
 	def viewHtml(self):
 		htmlDlg = HtmlDialog(self)
 		try:
-			htmltext = self.currentTab.getHtml(includeStyleSheet=False,
-				includeTitle=False)
+			_, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=False,
+			                                                      webenv=False)
 		except Exception:
 			return self.printError()
-		winTitle = self.currentTab.getDocumentTitle(baseName=True)
+		winTitle = self.currentTab.getBaseName()
 		htmlDlg.setWindowTitle(winTitle+" ("+self.tr("HTML code")+")")
 		htmlDlg.textEdit.setPlainText(htmltext.rstrip())
 		htmlDlg.hl.rehighlight()
@@ -1067,7 +1103,7 @@ class ReTextWindow(QMainWindow):
 		'<p><b>' + (self.tr('ReText %s (using PyMarkups %s)') % (app_version, markups.__version__))
 		+'</b></p>' + self.tr('Simple but powerful editor'
 		' for Markdown and reStructuredText')
-		+'</p><p>'+self.tr('Author: Dmitry Shachnev, 2011').replace('2011', '2011\u2013' '2015')
+		+'</p><p>'+self.tr('Author: Dmitry Shachnev, 2011').replace('2011', '2011–2016')
 		+'<br><a href="https://github.com/retext-project/retext">'+self.tr('Website')
 		+'</a> | <a href="http://daringfireball.net/projects/markdown/syntax">'
 		+self.tr('Markdown syntax')
@@ -1080,5 +1116,4 @@ class ReTextWindow(QMainWindow):
 		writeToSettings('defaultMarkup', markupClass.name, defaultName)
 		for tab in self.iterateTabs():
 			if not tab.fileName:
-				tab.setMarkupClass(markupClass)
-				tab.updatePreviewBox()
+				tab.setDefaultMarkupClass(markupClass)
